#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ido.h"



__global__  void  cuda_diffusion2d_0
// ====================================================================
//
// program    :  CUDA device code for 2-D diffusion equation
//               for 16 x 16 block and 16 x 16 thread per 1 block
//
// date       :  Nov 07, 2008
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *f,         /* dependent variable                        */
   float    *fn,        /* dependent variable                        */
   int      nx,         /* grid number in the x-direction            */
   int      ny,         /* grid number in the x-direction            */
   float    c0,         /* coefficient no.0                          */
   float    c1,         /* coefficient no.1                          */
   float    c2          /* coefficient no.2                          */
)
// --------------------------------------------------------------------
{
   int    j,    jx,   jy;
   float  fcc,  fce,  fcw,  fcs,  fcn;

   jy = blockDim.y*blockIdx.y + threadIdx.y;
   jx = blockDim.x*blockIdx.x + threadIdx.x;
   j = nx*jy + jx;

   fcc = f[j];

   if(jx == 0) fcw = fcc;
   else        fcw = f[j - 1];

   if(jx == nx - 1) fce = fcc;
   else             fce = f[j+1];

   if(jy == 0) fcs = fcc;
   else        fcs = f[j-nx];

   if(jy == ny - 1) fcn = fcc;
   else             fcn = f[j+nx];

   fn[j] = c0*(fce + fcw)
         + c1*(fcn + fcs)
         + c2*fcc;
}



void  cpu_diffusion2d
// ====================================================================
//
// purpos     :  2-dimensional diffusion equation solved by FDM
//
// date       :  May 16, 2008
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   float    *f,         /* dependent variable                        */
   float    *fn,        /* updated dependent variable                */
   int      nx,         /* x-dimensional grid size                   */
   int      ny,         /* y-dimensional grid size                   */
   float    c0,         /* coefficient no.0                          */
   float    c1,         /* coefficient no.1                          */
   float    c2          /* coefficient no.2                          */
)
// --------------------------------------------------------------------
{
     int    j,    jx,   jy;
     float  fcc,  fce,  fcw,  fcs,  fcn;

     for(jy = 0; jy < ny; jy++) {
         for(jx = 0; jx < nx; jx++) {
             j = nx*jy + jx;
             fcc = f[j];

             if(jx == 0) fcw = fcc;
             else        fcw = f[j - 1];

             if(jx == nx - 1) fce = fcc;
             else             fce = f[j+1];

             if(jy == 0) fcs = fcc;
             else        fcs = f[j-nx];

             if(jy == ny - 1) fcn = fcc;
             else             fcn = f[j+nx];

             fn[j] = c0*(fce + fcw)
                   + c1*(fcn + fcs)
                   + c2*fcc;
         }
     }
}



float  diffusion2d
// ====================================================================
//
// purpos     :  2-dimensional diffusion equation solved by FDM
//
// date       :  May 16, 2008
// programmer :  Takayuki Aoki
// place      :  Tokyo Institute of Technology
//
(
   int      numGPUs,    /* available number of GPU device            */
   int      nx,         /* x-dimensional grid size                   */
   int      ny,         /* y-dimensional grid size                   */
   float    *f,         /* dependent variable                        */
   float    *fn,        /* updated dependent variable                */
   float    kappa,      /* diffusion coefficient                     */
   float    dt,         /* time step interval                        */
   float    dx,         /* grid spacing in the x-direction           */
   float    dy          /* grid spacing in the y-direction           */
)
// --------------------------------------------------------------------
{
     float     c0 = kappa*dt/(dx*dx),   c1 = kappa*dt/(dy*dy),
               c2 = 1.0 - 2.0*(c0 + c1);

     if(numGPUs > 0) {

#define  blockDim_x    256
#define  blockDim_y    1

        dim3  grid(nx/blockDim_x,ny/blockDim_y,1),  threads(blockDim_x,blockDim_y,1);
        cuda_diffusion2d_0<<< grid, threads >>>(f,fn,nx,ny,c0,c1,c2);

        cudaThreadSynchronize();
     }
     else cpu_diffusion2d(f,fn,nx,ny,c0,c1,c2);

     return (float)(nx*ny)*7.0;
}
